import tkinter as tk
from tkinter import filedialog, messagebox, ttk
import cv2
import numpy as np
from PIL import Image, ImageTk, ImageGrab
import math

class StarRecognition:
    def __init__(self):
        self.root = tk.Tk()
        self.root.title("Распознавание звезд полярного полюса")
        self.root.geometry("1200x800")
        
        # Координаты звезд в градусах
        self.star_coords = {
            'polaris': {'ra': 37.5756, 'dec': 89.2642},    # 2h30m21.61s +89°15'41.9"
            'hip7283': {'ra': 23.2554, 'dec': 89.0116},    # 1h33m01.29s +89°00'41.9"
            'hip3128': {'ra': 9.6116, 'dec': 89.4397}      # 0h38m26.77s +89°26'22.8"
        }
        
        # Состояние программы
        self.image = None
        self.photo = None
        self.original_image = None
        self.selected_stars = {}  # {'polaris': (x, y), 'hip7283': (x, y), 'hip3128': (x, y)}
        self.current_star = 'polaris'
        self.star_order = ['polaris', 'hip7283', 'hip3128']
        self.star_names = {
            'polaris': 'Polaris',
            'hip7283': 'HIP 7283',
            'hip3128': 'HIP 3128'
        }
        
        # Параметры отображения
        self.zoom_factor = 1.0
        self.pan_x = 0
        self.pan_y = 0
        self.dragging = False
        self.drag_start_x = 0
        self.drag_start_y = 0
        
        self.setup_ui()
        
    def setup_ui(self):
        # Главный фрейм
        main_frame = tk.Frame(self.root)
        main_frame.pack(fill=tk.BOTH, expand=True)
        
        # Левое меню
        self.menu_frame = tk.Frame(main_frame, width=200, bg='lightgray')
        self.menu_frame.pack(side=tk.LEFT, fill=tk.Y, padx=5, pady=5)
        self.menu_frame.pack_propagate(False)
        
        # Кнопки меню
        tk.Label(self.menu_frame, text="Управление", bg='lightgray', font=('Arial', 12, 'bold')).pack(pady=10)
        
        tk.Button(self.menu_frame, text="Загрузить новое\nизображение", command=self.load_image_dialog, 
                 width=15, height=2).pack(pady=5)
        
        tk.Label(self.menu_frame, text="Масштаб:", bg='lightgray').pack(pady=(15,5))
        tk.Button(self.menu_frame, text="Увеличить (+)", command=self.zoom_in, width=15).pack(pady=2)
        tk.Button(self.menu_frame, text="Уменьшить (-)", command=self.zoom_out, width=15).pack(pady=2)
        tk.Button(self.menu_frame, text="По размеру окна", command=self.fit_to_window, width=15).pack(pady=2)
        
        tk.Button(self.menu_frame, text="Сбросить метки", command=self.reset_markers, 
                 width=15, height=2).pack(pady=15)
        
        # Статус
        self.status_label = tk.Label(self.menu_frame, text="", bg='lightgray', wraplength=180, justify=tk.LEFT)
        self.status_label.pack(pady=10)
        
        # Область изображения
        self.image_frame = tk.Frame(main_frame, bg='black')
        self.image_frame.pack(side=tk.RIGHT, fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        # Canvas для изображения
        self.canvas = tk.Canvas(self.image_frame, bg='black', cursor='crosshair')
        self.canvas.pack(fill=tk.BOTH, expand=True)
        
        # Привязка событий
        self.canvas.bind("<Button-1>", self.on_click)
        self.canvas.bind("<B1-Motion>", self.on_drag)
        self.canvas.bind("<ButtonRelease-1>", self.on_release)
        self.root.bind("<Control-v>", self.paste_from_clipboard)
        
        self.update_status()
        
    def load_image_dialog(self):
        file_path = filedialog.askopenfilename(
            title="Выберите изображение",
            filetypes=[("Изображения", "*.png *.jpg *.jpeg"), ("PNG", "*.png"), ("JPEG", "*.jpg *.jpeg")]
        )
        if file_path:
            self.load_image(file_path)
            
    def paste_from_clipboard(self, event):
        try:
            image = ImageGrab.grabclipboard()
            if image:
                self.original_image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
                self.reset_view()
                self.display_image()
        except Exception as e:
            pass
            
    def load_image(self, file_path):
        try:
            self.original_image = cv2.imread(file_path)
            if self.original_image is not None:
                self.reset_view()
                self.display_image()
        except Exception as e:
            pass
            
    def reset_view(self):
        self.selected_stars = {}
        self.current_star = 'polaris'
        self.zoom_factor = 1.0
        self.pan_x = 0
        self.pan_y = 0
        self.fit_to_window()
        self.update_status()
        
    def fit_to_window(self):
        if self.original_image is None:
            return
            
        canvas_width = self.canvas.winfo_width()
        canvas_height = self.canvas.winfo_height()
        
        if canvas_width <= 1 or canvas_height <= 1:
            self.root.after(100, self.fit_to_window)
            return
            
        img_height, img_width = self.original_image.shape[:2]
        
        zoom_x = canvas_width / img_width
        zoom_y = canvas_height / img_height
        self.zoom_factor = min(zoom_x, zoom_y) * 0.9
        
        self.pan_x = 0
        self.pan_y = 0
        self.display_image()
        
    def zoom_in(self):
        self.zoom_factor *= 1.2
        self.display_image()
        
    def zoom_out(self):
        self.zoom_factor /= 1.2
        self.display_image()
        
    def display_image(self):
        if self.original_image is None:
            return
            
        # Применяем масштабирование
        img_height, img_width = self.original_image.shape[:2]
        new_width = int(img_width * self.zoom_factor)
        new_height = int(img_height * self.zoom_factor)
        
        resized_image = cv2.resize(self.original_image, (new_width, new_height))
        
        # Применяем панорамирование
        canvas_width = self.canvas.winfo_width()
        canvas_height = self.canvas.winfo_height()
        
        start_x = max(0, -self.pan_x)
        start_y = max(0, -self.pan_y)
        end_x = min(new_width, canvas_width - self.pan_x)
        end_y = min(new_height, canvas_height - self.pan_y)
        
        if start_x < end_x and start_y < end_y:
            visible_image = resized_image[start_y:end_y, start_x:end_x]
            
            # Рисуем координатную сетку и звезды
            self.draw_overlays(visible_image, start_x, start_y)
            
            # Конвертируем для Tkinter
            image_rgb = cv2.cvtColor(visible_image, cv2.COLOR_BGR2RGB)
            pil_image = Image.fromarray(image_rgb)
            self.photo = ImageTk.PhotoImage(pil_image)
            
            self.canvas.delete("all")
            display_x = max(0, self.pan_x)
            display_y = max(0, self.pan_y)
            self.canvas.create_image(display_x, display_y, anchor=tk.NW, image=self.photo)
            
    def draw_overlays(self, image, offset_x, offset_y):
        # Рисуем метки выбранных звезд
        for star_name, (x, y) in self.selected_stars.items():
            # Пересчитываем координаты с учетом масштаба и смещения
            display_x = int(x * self.zoom_factor - offset_x)
            display_y = int(y * self.zoom_factor - offset_y)
            
            if 0 <= display_x < image.shape[1] and 0 <= display_y < image.shape[0]:
                cv2.circle(image, (display_x, display_y), 20, (0, 255, 0), 2)
                cv2.putText(image, self.star_names[star_name], 
                           (display_x + 25, display_y - 10),
                           cv2.FONT_HERSHEY_SIMPLEX, 0.6, (0, 255, 0), 2)
        
        # Рисуем координатную сетку, если выбраны все звезды
        if len(self.selected_stars) == 3:
            self.draw_coordinate_grid(image, offset_x, offset_y)
            
    def draw_coordinate_grid(self, image, offset_x, offset_y):
        # Вычисляем положение полюса
        pole_x, pole_y = self.calculate_pole_position()
        
        if pole_x is None or pole_y is None:
            return
            
        # Пересчитываем координаты полюса с учетом масштаба и смещения
        display_pole_x = int(pole_x * self.zoom_factor - offset_x)
        display_pole_y = int(pole_y * self.zoom_factor - offset_y)
        
        # Рисуем полюс желтой точкой
        if (-20 <= display_pole_x < image.shape[1] + 20 and 
            -20 <= display_pole_y < image.shape[0] + 20):
            cv2.circle(image, (display_pole_x, display_pole_y), 6, (0, 255, 255), -1)
            cv2.circle(image, (display_pole_x, display_pole_y), 10, (0, 255, 255), 2)
            cv2.putText(image, "NCP", (display_pole_x + 15, display_pole_y - 15),
                       cv2.FONT_HERSHEY_SIMPLEX, 0.6, (0, 255, 255), 2)
        
        # Вычисляем масштаб для сетки на основе расстояния между звездами
        polaris_pos = self.selected_stars.get('polaris')
        hip7283_pos = self.selected_stars.get('hip7283')
        
        if polaris_pos and hip7283_pos:
            # Расстояние между полярной и HIP 7283 на изображении
            img_distance = math.sqrt((polaris_pos[0] - hip7283_pos[0])**2 + 
                                   (polaris_pos[1] - hip7283_pos[1])**2)
            
            # Теоретическое угловое расстояние между этими звездами (~0.26°)
            theoretical_angle = 0.26  # градусов
            pixels_per_degree = img_distance / theoretical_angle if theoretical_angle > 0 else 100
        else:
            pixels_per_degree = 100
        
        # Рисуем концентрические окружности для склонений
        center_x, center_y = display_pole_x, display_pole_y
        
        # Окружности для склонений: каждый градус от полюса
        for dec_offset in [1, 2, 3, 4, 5]:  # 1°, 2°, 3°, 4°, 5° от полюса
            radius_pix = int(dec_offset * pixels_per_degree * self.zoom_factor)
            if 10 < radius_pix < min(image.shape) * 2:
                cv2.circle(image, (center_x, center_y), radius_pix, (0, 0, 255), 1)
                # Подпись склонения
                label_x = center_x + int(radius_pix * 0.707)  # cos(45°)
                label_y = center_y - int(radius_pix * 0.707)  # sin(45°)
                if (0 < label_x < image.shape[1] - 40 and 
                    0 < label_y < image.shape[0] - 15):
                    cv2.putText(image, f"{90-dec_offset}°", 
                               (label_x, label_y),
                               cv2.FONT_HERSHEY_SIMPLEX, 0.4, (0, 0, 255), 1)
        
        # Рисуем линии прямого восхождения (каждые 2 часа = 30°)
        max_radius = int(5 * pixels_per_degree * self.zoom_factor)
        
        for hour in range(0, 24, 2):
            angle_deg = hour * 15  # 1 час = 15°
            # Поворачиваем на 90° чтобы 0h было вверху
            angle_rad = math.radians(angle_deg - 90)
            
            # Линия от центра
            end_x = int(center_x + max_radius * math.cos(angle_rad))
            end_y = int(center_y + max_radius * math.sin(angle_rad))
            
            cv2.line(image, (center_x, center_y), (end_x, end_y), (0, 0, 255), 1)
            
            # Подпись часов RA
            label_radius = max_radius * 0.8
            label_x = int(center_x + label_radius * math.cos(angle_rad))
            label_y = int(center_y + label_radius * math.sin(angle_rad))
            
            if (0 < label_x < image.shape[1] - 30 and 
                0 < label_y < image.shape[0] - 15):
                cv2.putText(image, f"{hour}h", (label_x, label_y),
                           cv2.FONT_HERSHEY_SIMPLEX, 0.4, (0, 0, 255), 1)
            
    def calculate_pole_position(self):
        if len(self.selected_stars) < 3:
            return None, None
            
        # Получаем координаты выбранных звезд на изображении
        polaris_pos = self.selected_stars.get('polaris')
        hip7283_pos = self.selected_stars.get('hip7283')
        hip3128_pos = self.selected_stars.get('hip3128')
        
        if not all([polaris_pos, hip7283_pos, hip3128_pos]):
            return None, None
            
        # Экваториальные координаты звезд (в градусах)
        # RA переводим из часов в градусы: часы * 15
        stars_coords = {
            'polaris': {'ra': 37.6, 'dec': 89.264},     # 2h30m → 37.5°
            'hip7283': {'ra': 23.25, 'dec': 89.012},    # 1h33m → 23.25°
            'hip3128': {'ra': 9.61, 'dec': 89.440}      # 0h38m → 9.61°
        }
        
        # Переводим экваториальные координаты в декартовы координаты на единичной сфере
        def equatorial_to_cartesian(ra_deg, dec_deg):
            ra_rad = math.radians(ra_deg)
            dec_rad = math.radians(dec_deg)
            x = math.cos(dec_rad) * math.cos(ra_rad)
            y = math.cos(dec_rad) * math.sin(ra_rad)
            z = math.sin(dec_rad)
            return np.array([x, y, z])
        
        # Декартовы координаты звезд на небесной сфере
        polaris_sky = equatorial_to_cartesian(stars_coords['polaris']['ra'], stars_coords['polaris']['dec'])
        hip7283_sky = equatorial_to_cartesian(stars_coords['hip7283']['ra'], stars_coords['hip7283']['dec'])
        hip3128_sky = equatorial_to_cartesian(stars_coords['hip3128']['ra'], stars_coords['hip3128']['dec'])
        
        # Координаты звезд на изображении
        polaris_img = np.array(polaris_pos)
        hip7283_img = np.array(hip7283_pos)
        hip3128_img = np.array(hip3128_pos)
        
        # Находим трансформацию методом наименьших квадратов
        # Создаем матрицы для системы уравнений
        sky_points = np.array([polaris_sky[:2], hip7283_sky[:2], hip3128_sky[:2]])  # Используем только x,y проекции
        img_points = np.array([polaris_img, hip7283_img, hip3128_img])
        
        # Находим аффинную трансформацию от небесных координат к координатам изображения
        # Добавляем единичный столбец для аффинной трансформации
        sky_homogeneous = np.column_stack([sky_points, np.ones(3)])
        
        try:
            # Решаем систему уравнений для трансформации
            transform_x = np.linalg.lstsq(sky_homogeneous, img_points[:, 0], rcond=None)[0]
            transform_y = np.linalg.lstsq(sky_homogeneous, img_points[:, 1], rcond=None)[0]
            
            # Северный полюс мира имеет координаты (0, 0, 1) в декартовой системе
            # или RA=любое, Dec=90° в экваториальной системе
            pole_sky = np.array([0, 0])  # Проекция полюса на плоскость xy
            pole_homogeneous = np.array([pole_sky[0], pole_sky[1], 1])
            
            # Применяем трансформацию для нахождения полюса на изображении
            pole_x = np.dot(transform_x, pole_homogeneous)
            pole_y = np.dot(transform_y, pole_homogeneous)
            
            return pole_x, pole_y
            
        except np.linalg.LinAlgError:
            # Если не удается решить систему, используем геометрический подход
            return self.calculate_pole_geometric()
    
    def calculate_pole_geometric(self):
        """Альтернативный геометрический расчет полюса"""
        polaris_pos = self.selected_stars.get('polaris')
        hip7283_pos = self.selected_stars.get('hip7283')
        hip3128_pos = self.selected_stars.get('hip3128')
        
        # Простой геометрический подход: 
        # Полюс находится в центре окружности, описанной вокруг звезд
        x1, y1 = polaris_pos
        x2, y2 = hip7283_pos
        x3, y3 = hip3128_pos
        
        # Находим центр описанной окружности
        d = 2 * (x1 * (y2 - y3) + x2 * (y3 - y1) + x3 * (y1 - y2))
        
        if abs(d) > 1e-10:
            ux = ((x1*x1 + y1*y1) * (y2 - y3) + (x2*x2 + y2*y2) * (y3 - y1) + (x3*x3 + y3*y3) * (y1 - y2)) / d
            uy = ((x1*x1 + y1*y1) * (x3 - x2) + (x2*x2 + y2*y2) * (x1 - x3) + (x3*x3 + y3*y3) * (x2 - x1)) / d
            return ux, uy
        else:
            # Если точки коллинеарны, используем центр масс
            return (x1 + x2 + x3) / 3, (y1 + y2 + y3) / 3
        
    def find_centroid(self, x, y, radius=15):
        """Находит центроид светового пятна звезды"""
        if self.original_image is None:
            return x, y
            
        h, w = self.original_image.shape[:2]
        
        # Преобразуем координаты клика в координаты исходного изображения
        orig_x = int(x / self.zoom_factor)
        orig_y = int(y / self.zoom_factor)
        
        # Определяем область поиска
        x1 = max(0, orig_x - radius)
        y1 = max(0, orig_y - radius)
        x2 = min(w, orig_x + radius)
        y2 = min(h, orig_y + radius)
        
        # Извлекаем область изображения
        roi = self.original_image[y1:y2, x1:x2]
        gray_roi = cv2.cvtColor(roi, cv2.COLOR_BGR2GRAY)
        
        # Находим центроид
        moments = cv2.moments(gray_roi)
        if moments['m00'] != 0:
            cx = moments['m10'] / moments['m00']
            cy = moments['m01'] / moments['m00']
            
            # Преобразуем обратно в координаты исходного изображения
            centroid_x = x1 + cx
            centroid_y = y1 + cy
            
            return centroid_x, centroid_y
        
        return orig_x, orig_y
        
    def on_click(self, event):
        if self.original_image is None:
            return
            
        if len(self.selected_stars) >= 3:
            self.dragging = True
            self.drag_start_x = event.x
            self.drag_start_y = event.y
            return
            
        # Преобразуем координаты клика с учетом панорамирования
        click_x = event.x - self.pan_x
        click_y = event.y - self.pan_y
        
        # Находим центроид звезды
        centroid_x, centroid_y = self.find_centroid(click_x, click_y)
        
        # Сохраняем координаты текущей звезды
        self.selected_stars[self.current_star] = (centroid_x, centroid_y)
        
        # Переходим к следующей звезде
        current_index = self.star_order.index(self.current_star)
        if current_index < len(self.star_order) - 1:
            self.current_star = self.star_order[current_index + 1]
        
        self.display_image()
        self.update_status()
            
    def on_drag(self, event):
        if self.dragging and len(self.selected_stars) >= 3:
            dx = event.x - self.drag_start_x
            dy = event.y - self.drag_start_y
            self.pan_x += dx
            self.pan_y += dy
            self.drag_start_x = event.x
            self.drag_start_y = event.y
            self.display_image()
            
    def on_release(self, event):
        self.dragging = False
        
    def reset_markers(self):
        self.selected_stars = {}
        self.current_star = 'polaris'
        self.display_image()
        self.update_status()
        
    def update_status(self):
        if len(self.selected_stars) == 0:
            status = f"Отметьте {self.star_names[self.current_star]}"
        elif len(self.selected_stars) < 3:
            status = f"Отмечено: {len(self.selected_stars)}/3\nСледующая: {self.star_names[self.current_star]}"
        else:
            status = "Все звезды отмечены!\nМожно перемещать\nизображение мышью"
        
        self.status_label.config(text=status)
        
    def run(self):
        self.root.mainloop()

if __name__ == "__main__":
    app = StarRecognition()
    app.run()